"""Custom UMAP function that adds labels to the plot."""

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def umap(
    adata,
    color=None,
    add_labels=True,
    key_umap="X_umap",
    cmap="tab20",
    ax=None,
    remove_legend=True,
    remove_axes=True,
    kwargs_scatter={},
    kwargs_text={},
):
    """Plot UMAP and add labels to the plot."""
    # defaults
    default_scatter = {
        "linewidth": 0,
        "s": 5,
    }
    kwargs_scatter = {**default_scatter, **kwargs_scatter}

    # bold and white shadow and darken a bit
    defaults_text = {
        "fontsize": 10,
        "ha": "center",
        "va": "center",
        "fontweight": "bold",
        "bbox": dict(
            facecolor="black", alpha=0.5, edgecolor="none", boxstyle="round,pad=0.3"
        ),
    }
    kwargs_text = {**defaults_text, **kwargs_text}

    # assertions
    assert isinstance(color, str), "color must a column name in adata.obs"
    assert color in adata.obs.columns, f"{color} not in adata.obs"
    assert key_umap in adata.obsm.keys(), f"{key_umap} not in adata.obsm"
    assert adata.obsm[key_umap].shape[1] == 2, f"{key_umap} must have 2 dimensions"

    # only categorical supported at this point
    supported = [pd.CategoricalDtype, pd.StringDtype]
    assert any([isinstance(adata.obs[color].dtype, dtype) for dtype in supported]), (
        f"{color} must be one of {supported} types"
    )

    # prepare data
    df = pd.DataFrame(
        {
            "x": adata.obsm[key_umap][:, 0],
            "y": adata.obsm[key_umap][:, 1],
            "color": adata.obs[color],
        }
    )

    # prepare cmap
    labels = df["color"].unique()
    if isinstance(cmap, str):
        cmap = dict(zip(labels, sns.color_palette(cmap, len(labels))))
    elif isinstance(cmap, dict):
        assert all([label in cmap.keys() for label in labels]), (
            f"cmap must have all labels in {labels}. Missing: {set(labels) - set(cmap.keys())}"
        )

    # prepare label locations (cluster centroids)
    df_labels = df.groupby("color", observed=True).mean().reset_index()

    # prepare fig, ax
    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 8))

    # plot
    ax = sns.scatterplot(
        df,
        x="x",
        y="y",
        hue="color",
        palette=cmap,
        ax=ax,
        **kwargs_scatter,
    )

    # add labels
    if add_labels:
        for i, row in df_labels.iterrows():
            kwargs_text["bbox"]["facecolor"] = [c * 0.9 for c in cmap[row["color"]]]
            ax.text(
                row["x"],
                row["y"],
                row["color"],
                color="white",
                **kwargs_text,
            )

    # remove legend (it's overlayed anyway)
    if remove_legend:
        ax.legend().remove()

    # remove axes (not informative in umaps)
    if remove_axes:
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

    return ax



"""UMI rankplots, analogous to 10X web summaries."""

import math

import anndata as ad
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.cm import ScalarMappable
from matplotlib.colors import LinearSegmentedColormap, Normalize

from .._defaults import DEFAULTS
from .._logging import get_logger
from .._utils import get_layer, to_dense_safe


def scale_cmap(color_min, color_max, vmin=0, vmax=1):
    """Scale colormap between 2 colors."""
    cmap = LinearSegmentedColormap.from_list("custom", [color_max, color_min])
    return ScalarMappable(norm=Normalize(vmin=vmin, vmax=vmax), cmap=cmap)


def bucketize(
    df: pd.DataFrame,
    key_counts: str = "counts",
    key_values: str = "_values_temp",
    n_buckets=150,
):
    """Aggregate buckets based on counts. Group if provided. Aggregate mean if values provided."""
    # initialize
    df = df.copy()

    if key_values == "_values_temp":
        df.loc[:, key_values] = 1
    assert df[key_values].min() >= 0 and df[key_values].max() <= 1, (
        "Values must be between 0 and 1"
    )

    # add rank based on counts, descending
    df.loc[:, "_rank"] = np.log10(df[key_counts].rank(ascending=False) + 1)
    cuts = np.linspace(df["_rank"].min(), df["_rank"].max(), n_buckets + 1)
    labels = cuts[1:]
    df.loc[:, "_bucket"] = pd.cut(
        df["_rank"], bins=cuts, labels=labels, include_lowest=True
    ).astype(float)
    df.loc[:, "_bucket"] = df["_bucket"].round(2)

    # aggregate:
    # - mean 'key_counts'
    # - mean 'key_values'
    df_agg = (
        df.groupby(["_bucket"], observed=True)
        .agg(
            **{
                key_counts: (key_counts, "mean"),
                key_values: (key_values, "mean"),
                "cells": (key_counts, "size"),
            }
        )
        .reset_index()
    )

    # filter buckets with at least 4 cells
    df_agg = df_agg[df_agg["cells"] >= 4]

    # sort by bucket
    df_agg = df_agg.sort_values("_bucket")
    return df_agg


def umi_one(
    df: pd.DataFrame,
    ax: plt.Axes = None,
    use_log: bool = True,
    key_counts: str = "counts",
    key_values: str = "_values_temp",
    n_buckets=50,
    color: str = "#1f77b4",  # blue
    color_fade="#D3D3D3",  # fade to lightgrey by default
    verbose: int = DEFAULTS["verbose"],
    **kwargs,
):
    """Plot UMI counts for a single variable.

    Args:
        df (pd.DataFrame): DataFrame containing UMI counts.
        ax (plt.Axes, optional): Matplotlib axes. Defaults to None.
        use_log (bool, optional): Whether to use log scale. Defaults to True.
        key_counts (str, optional): Key for counts. Defaults to 'counts'.
        key_values (str, optional): Key for values. Defaults to '_values_temp'.
        n_buckets (int, optional): Number of buckets for aggregation. Defaults to 50.
        color (str, optional): Color for the plot. Defaults to '#1f77b4'.
        color_fade (str, optional): Color to fade to. Defaults to '#D3D3D3'.
        verbose (int, optional): Verbosity level. Defaults to DEFAULTS["verbose"].
        **kwargs: Additional keyword arguments.

    """
    df = df.copy()
    if use_log:
        df.loc[:, key_counts] = np.log10(df[key_counts].astype(float) + 1)

    logger = get_logger("plot", level=verbose)

    # prep data
    logger.debug(f"Used df: {df.shape}\n{df.head(10)}")
    df_agg = bucketize(
        df, key_counts=key_counts, key_values=key_values, n_buckets=n_buckets
    )
    logger.debug(
        f"Aggregated df: {df_agg.shape}\n{df_agg.sort_values('_bucket').head(10)}"
    )
    logger.debug(
        f"Aggregated df: {df_agg.shape}\n{df_agg.sort_values('_bucket').tail(10)}"
    )

    # set color scale
    if key_values == "_values_temp":
        palette = [color] * len(df_agg[key_values].unique())
    else:
        cmap_fades = scale_cmap(
            color_fade,
            color,
            vmin=df_agg[key_values].min(),
            vmax=df_agg[key_values].max(),
        )
        palette = [cmap_fades.to_rgba(i) for i in df_agg[key_values].unique()]

    ax = sns.scatterplot(
        data=df_agg,
        x="_bucket",
        y=key_counts,
        hue=key_values,
        palette=palette,
        ax=ax,
        **kwargs,
    )

    if use_log:
        ax = _add_log10_axis(ax, key_counts=key_counts)

    # rotate xticks by 0-90 degrees, and only show every 5th
    return ax


def _format_ticks(x):
    if x > 10:
        return f"{int(x):,}"
    elif x > 1:
        return f"{x:.1f}"
    else:
        return f"{x:.2f}"


def _add_log10_axis(ax: plt.Axes, key_counts: str = "counts"):
    """Add a log10 axis to the plot."""
    # add logged axis label
    ax.set_ylabel(f"Log10 {key_counts}")
    ax.set_xlabel(f"Log10 {key_counts} Rank")
    # xticks
    max_i = math.floor(ax.get_xlim()[1])
    xtick_positions = np.log10(
        np.array([10, 100, 1000, 10000, 100000, 1000000][:max_i]) + 1
    )
    xtick_labels = ["10", "100", "1k", "10k", "100k", "1M"][:max_i]
    ax.set_xticks(xtick_positions)
    ax.set_xticklabels(xtick_labels)
    # yticks
    yticks = ax.get_yticks()
    ax.set_yticks(yticks)
    ax.set_yticklabels([f"{_format_ticks(i)}" for i in yticks])
    return ax


def umi(
    adata: ad.AnnData,
    ax: plt.Axes = None,
    use_layer: str = DEFAULTS["use_layer"],
    key_counts: str = "counts",
    key_groups: str = "_groups_temp",
    key_values: str = "_values_temp",
    axis: int = 0,
    use_log: bool = True,
    each_var: bool = False,
    n_buckets=50,
    cmap=None,
    color_fade="#D3D3D3",  # fade to lightgrey by default
    verbose: int = DEFAULTS["verbose"],
    **kwargs,
):
    """Plot UMI counts for each HTO or other groups.

    Args:
        adata (ad.AnnData): AnnData object.
        ax (plt.Axes, optional): Matplotlib axes. Defaults to None.
        use_layer (str, optional): {DESCRIPTIONS["use_layer"]}
        key_counts (str, optional): Key for counts. Defaults to 'counts'.
        key_groups (str, optional): Key for groups. Defaults to '_groups_temp'.
        key_values (str, optional): Key for values. Defaults to '_values_temp'. Recommended to point to a column of 0-1 values, such as "filtered" cells.
        use_log (bool, optional): Use log scale. Defaults to True.
        each_var (bool, optional): Plot each column of the layer, typically each HTO. Defaults to False.
        n_buckets (int, optional): Number of buckets. Defaults to 30.
        cmap (str, optional): Color map. Defaults to None.
        color_fade (str, optional): Fade to a color. Defaults to '#D3D3D3'.
        verbose (int, optional): {DESCRIPTIONS["verbose"]}
        axis (int, optional): Axis to plot. Defaults to 0.
        **kwargs: Additional keyword arguments used in sns.scatterplot.

    """
    # init
    if ax is None:
        fig, ax = plt.subplots(1, 1)

    # get data
    if axis == 0:
        df = adata.obs.copy()
        _, X = get_layer(adata, use_layer, numpy=False, inplace=False)
    elif axis == 1:
        df = adata.var.copy()
        _, X = get_layer(adata, use_layer, numpy=False, inplace=False)
        X = X.T  # transpose to have genes in rows
    else:
        raise ValueError(f"Axis must be 0 or 1, got {axis}")

    # concatenate if each_ver
    if each_var:
        assert adata.shape[1] < 19, (
            f"Number of vars must be less than 19, got {adata.shape[1]}"
        )
        assert not key_groups != "_groups_temp", (
            "Can't use 'each_var' with 'key_groups', as variable names are used as groups"
        )
        key_groups = "var_name"
        df = pd.concat([df] * adata.shape[1], ignore_index=True)
        df.loc[:, key_groups] = np.repeat(adata.var_names, adata.shape[0])
        df.loc[:, key_counts] = np.array(to_dense_safe(X.T)).flatten()
    else:
        # sum
        df.loc[:, key_counts] = np.asarray(X.sum(axis=1), dtype=np.float64).flatten()

    # plot
    if key_groups == "_groups_temp":
        # set color
        if cmap is None:
            cmap = "#1f77b4"

        # plot
        ax = umi_one(
            df,
            ax=ax,
            use_log=use_log,
            key_counts=key_counts,
            key_values=key_values,
            n_buckets=n_buckets,
            color=cmap,
            color_fade=color_fade,
            verbose=verbose,
            **kwargs,
        )
        ax.get_legend().remove()
    else:
        # set colors
        groups = df[key_groups].unique()
        if cmap is None:
            cmap = "tab10"
        if isinstance(cmap, str):
            cmap = dict(zip(groups, sns.color_palette(cmap)))

        # plot each group
        for j, group in enumerate(groups):
            # get
            df_temp = df[df[key_groups] == group].copy()

            ax = umi_one(
                df_temp,
                ax=ax,
                use_log=use_log,
                key_counts=key_counts,
                key_values=key_values,
                n_buckets=n_buckets,
                color=cmap[group],
                color_fade=color_fade,
                verbose=verbose,
                **kwargs,
            )

        # add legend
        ax.get_legend().remove()
        handles = [
            plt.Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                markerfacecolor=cmap[group],
                label=group,
            )
            for group in groups
        ]
        ax.legend(handles=handles, title=key_groups, loc="lower left")

    ax.set_xlabel("Log UMI Rank")
    return ax



"""Create a simple HTO demultiplexing report."""

import os

import matplotlib.backends.backend_pdf
import matplotlib.pyplot as plt

import hto.pl as pl

from ._defaults import DEFAULTS
from ._logging import get_logger
from ._utils import get_arg, savepdf, subset_whitelist

REPORT_PLT_DEFAULTS = {"dpi": 80}


def report_safe(*args, **kwargs):
    """Run report with error handling."""
    try:
        report(*args, **kwargs)
    except Exception as e:
        logger = get_logger("report", level=get_arg("verbose", kwargs, DEFAULTS))
        logger.error(f"Failed to run report: {e}")
        if kwargs["verbose"] >= 2:
            raise e


def report(
    adata_hto,
    adata_background,
    adata_hto_raw,
    adata_gex,
    path_report: str = None,
    show: bool = False,
    use_key_normalise: str = DEFAULTS["add_key_normalise"],
    use_key_denoise: str = DEFAULTS["add_key_denoised"],
    verbose: int = DEFAULTS["verbose"],
):
    """Create a simple demultiplexing report.

    Generates:
    - umiplot
    - umi_gex_hto
    - distributions_stages
    - technical_noise

    This requires some specialised inputs. This function is best used as part of
    hto.demultiplex().
    """
    # setup
    logger = get_logger("report", level=verbose)

    if path_report is None:
        if not show:
            return
        show = True
        pdf = None
    else:
        assert path_report.endswith(".pdf"), "Path must end with .pdf"
        logger.info(f"Creating report at '{path_report}'")
        os.makedirs(os.path.dirname(path_report), exist_ok=True)
        pdf = matplotlib.backends.backend_pdf.PdfPages(path_report)

    # assertions
    if use_key_normalise is None or use_key_denoise is None:
        logger.error(
            "Skipping report. Require parameters 'add_key_normalise' (--add-key-normalise) "
            "and 'add_key_denoised' (--add-key-denoise) to generate report."
        )
        return

    # preprocess
    if adata_background is None:
        adata_background = subset_whitelist(
            adata_hto_raw, adata_hto.uns["dnd"]["normalise"]["params"]["background"]
        )
    else:
        adata_background = adata_background.copy()
    adata_background.obs.loc[:, "filtered"] = adata_background.obs_names.isin(
        adata_hto.obs_names
    )

    # plot umiplot
    fig_umiplot, ax = plt.subplots(1, 1, figsize=(8, 4), **REPORT_PLT_DEFAULTS)

    ax = pl.umi(
        adata_background,
        key_values="filtered",
        each_var=True,
        verbose=1,
        ax=ax,
    )
    savepdf(pdf=pdf, path=path_report, fig=fig_umiplot, show=show)

    # umi_gex_hto

    if adata_gex is not None:
        fig_umi_gex_hto, axs = plt.subplots(1, 2, figsize=(8, 4), **REPORT_PLT_DEFAULTS)
        _ = pl.umi_gex_hto(
            adata_hto=adata_hto,
            adata_background=adata_background,
            adata_hto_raw=adata_hto_raw,
            adata_gex=adata_gex,
            axs=axs,
        )
        savepdf(pdf=pdf, path=path_report, fig=fig_umi_gex_hto, show=show)
    else:
        logger.info("Skipping umi_gex_hto plot as adata_gex is not provided")

    # distributions_stages
    fig_distributions_stages, axs = plt.subplots(
        3, 1, figsize=(8, 12), **REPORT_PLT_DEFAULTS
    )

    axs = pl.distribution_stages(
        adata=adata_hto,
        use_key_normalise=use_key_normalise,
        use_key_denoise=use_key_denoise,
        axs=axs,
    )
    savepdf(pdf=pdf, path=path_report, fig=fig_distributions_stages, show=show)

    # technical_noise
    for i in range(adata_hto.shape[1]):
        fig_technical_noise, axs = plt.subplots(
            2,
            2,
            figsize=(10, 10),
            gridspec_kw={"width_ratios": [3, 1]},
            sharey="row",
            **REPORT_PLT_DEFAULTS,
        )

        axs = pl.technical_noise(
            adata=adata_hto,
            var=i,
            use_key_normalise=use_key_normalise,
            use_key_denoise=use_key_denoise,
            axs=axs,
        )
        savepdf(pdf=pdf, path=path_report, fig=fig_technical_noise, show=show)

    # close
    if pdf is not None:
        pdf.close()

    return None



"""Generate a CLI command using Click. Arguments are automatically generated from `hto._defaults.py` options."""

import click

from hto import demultiplex
from hto._defaults import OPTIONS


@click.group(
    help="""\
CLI for HTO demultiplexing and normalisation.

Example run:

\b
hto demultiplex \\
    --adata-hto /path/to/adata_hto.h5ad \\
    --adata-hto-raw /path/to/adata_hto_raw.h5ad \\
    --adata-gex /path/to/adata_gex.h5ad \\
    --adata-out /path/to/output.h5ad

For all options, run:
\b
hto demultiplex --help
"""
)
def cli():
    """CLI for HTO demultiplexing and normalisation."""
    pass


# create function
@cli.command(
    name="demultiplex",
    help="Perform normalisation and demultiplexing of HTO data.",
    no_args_is_help=True,
)
def demultiplex_cli(**kwargs):
    """Run demultiplexing and normalization of HTO data."""
    demultiplex(_as_cli=True, **kwargs)


# add options
for key, option in OPTIONS.items():
    # skip anonymous options
    if key[0] == "_":
        continue
    demultiplex_cli = option(demultiplex_cli)

if __name__ == "__main__":
    cli()



"""Introducing two metrics: Singlet Accuracy (SA) and Singlet Recovery Rate (SRR)."""

import numpy as np
from sklearn.metrics import f1_score

# y_true: ground truth labels
# y_pred: predicted labels from your algorithm


def singlet_precision(
    ground_truth,
    prediction,
    non_singlet_substring: list = ["low_quality", "doublet", ":", "negative"],
    ignore_case: bool = True,
):
    """Singlet Precision (SP) measures the percentage of predicted singlets that are correctly classified.

    Args:
        ground_truth (pd.Series): Ground truth labels.
        prediction (pd.Series): Predicted labels.
        non_singlet_substring (list): List of substrings that indicate a non-singlet label.
        ignore_case (bool): Whether to ignore case when checking for non-singlet substrings.

    """
    # ignore case
    if ignore_case:
        prediction = prediction.str.lower()
        ground_truth = ground_truth.str.lower()
        non_singlet_substring = [sub.lower() for sub in non_singlet_substring]

    # subset predicted singlets
    is_predicted_singlet = np.all(
        [~prediction.str.contains(sub) for sub in non_singlet_substring], axis=0
    )
    predicted_singlets = prediction[is_predicted_singlet]
    ground_truth_singlets = ground_truth[is_predicted_singlet]

    # return precision
    precision = (predicted_singlets == ground_truth_singlets).mean()
    return precision


def singlet_sensitivity(
    ground_truth,
    prediction,
    non_singlet_substring: list = ["low_quality", "doublet", ":", "negative"],
    ignore_case: bool = True,
):
    """Singlet Sensitivity measures the percentage of ground truth singlets that are correctly predicted.

    Args:
        ground_truth (pd.Series): Ground truth labels.
        prediction (pd.Series): Predicted labels.
        non_singlet_substring (list): List of substrings that indicate a non-singlet label.
        ignore_case (bool): Whether to ignore case when checking for non-singlet substrings.

    """
    # ignore case
    if ignore_case:
        prediction = prediction.str.lower()
        ground_truth = ground_truth.str.lower()
        non_singlet_substring = [sub.lower() for sub in non_singlet_substring]

    # subset ground truth singlets
    is_true_singlet = np.all(
        [~ground_truth.str.contains(sub) for sub in non_singlet_substring], axis=0
    )
    predicted_singlets = prediction[is_true_singlet]
    ground_truth_singlets = ground_truth[is_true_singlet]

    # return sensitivity
    sensitivity = (predicted_singlets == ground_truth_singlets).mean()
    return sensitivity


def get_metrics(ground_truth, prediction):
    """Obtain metrices of a ground_truth and prediction.

    - SP (Singlet Precision) = Correctly Labelled Predicted Singlets / All Predicted Singlets
    - SS (Singlet Sensitivity) = Correctly Labelled Ground Truth Singlets / All Ground Truth Singlets
    """
    return {
        "singlet_sensitivity": singlet_sensitivity(ground_truth, prediction),
        "singlet_precision": singlet_precision(ground_truth, prediction),
        "f1": f1_score(ground_truth, prediction, average="micro"),
    }



"""HTO Package."""

from . import data, metrics, pl, tl
from .demux import demux
from .denoise import denoise
from .dnd import demultiplex
from .normalise import normalise, normalise_debug

__all__ = [
    "normalise",
    "normalise_debug",
    "denoise",
    "demux",
    "metrics",
    "demultiplex",
    "tl",
    "data",
    "pl",
]



"""Visualise the distribution of HTO expression values."""

import anndata as ad
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import seaborn as sns

from .._defaults import DEFAULTS


def _set_lims(ax, xmin=1):
    ### BROKEN
    # if xmin is None:
    #    return ax

    cutoff_y = 0
    lines = ax.get_lines()

    # ensure to include all ylims above xmin, but still cutoff at 99th percentile
    for i in range(len(lines)):
        x_data, y_data = lines[i].get_data()
        y_sub = y_data[x_data > xmin]
        if len(y_sub) == 0:
            continue
        cutoff_y = max(cutoff_y, np.quantile(y_sub, 0.9999))
    ax.set_ylim(0, cutoff_y * 1.3)
    return ax


def _symmetric_log1p(x):
    return np.sign(x) * np.log1p(np.abs(x))


def _format(x):
    if np.expm1(x) < 1000:
        return f"{np.expm1(x):.0f}"
    elif np.expm1(x) < 1000000:
        return f"{np.expm1(x) / 1000:.0f}K"
    else:
        return f"{np.expm1(x) / 1000000:.0f}M"


def distribution(
    adata,
    ax=None,
    layer=None,
    cmap="tab20",
    title="",
    remove_legend=False,
    params_legend={},
    use_log=True,
    **kwargs,
):
    """Plot the distribution of HTO expression data.

    Args:
        adata (AnnData): The annotated data object containing the HTO expression data.
        ax (plt.Axes, optional): The axes to plot on. If None, a new figure is created.
        layer (str, optional): The layer to use for the plot. Defaults to None.
        cmap (str, optional): The colormap to use for the plot. Defaults to "tab20".
        title (str, optional): The title of the plot. Defaults to "".
        remove_legend (bool, optional): If True, remove the legend from the plot. Defaults to False.
        params_legend (dict, optional): Additional parameters for the legend.
        use_log (bool, optional): If True, use a log scale for the x-axis. Defaults to True.
        **kwargs: Additional keyword arguments for the seaborn kdeplot.

    """
    # defaults
    defaults_legend = {
        "title": "Hashtags",
        "bbox_to_anchor": (1.05, 1),
        "loc": "upper left",
    }
    params_legend = {**defaults_legend, **params_legend}

    defaults_kdeplot = {
        "fill": True,
    }
    params_kdeplot = {**defaults_kdeplot, **kwargs}

    # prep data
    df_long = adata.to_df(layer).melt(var_name="variable", value_name="value")
    if use_log:
        df_long.loc[:, "value_set"] = _symmetric_log1p(df_long.value)
    else:
        df_long.loc[:, "value_set"] = df_long.value

    # plot
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 3))

    # kde log
    ax = sns.kdeplot(
        df_long, hue="variable", x="value_set", palette=cmap, ax=ax, **params_kdeplot
    )
    # ax = _set_lims(ax, xmin=xmin)
    ax.set_title(title)
    ax.yaxis.set_ticks([])
    ax.set_ylabel("")

    # log transform
    if use_log:
        ax.set_xlabel("Logged Antibody Count")
        log_ticks = _symmetric_log1p(
            [-100000, -1000, -10, -1, 0, 1, 10, 100, 1000, 10000, 100000]
        )  # Replace with dynamic range if needed
        ax.xaxis.set_major_locator(ticker.FixedLocator(log_ticks))
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, _: _format(x)))
    else:
        ax.set_xlabel("Antibody Count")

    if remove_legend:
        ax.get_legend().remove()
    else:
        handles = ax.get_legend().legend_handles
        labels = [t._text for t in ax.get_legend().texts]
        ax.legend(handles, labels, **params_legend)

    return ax


def distribution_stages(
    adata: ad.AnnData,
    figsize=(8, 12),
    layer_raw=None,
    highlight: int = None,
    use_key_normalise=DEFAULTS["add_key_normalise"],
    use_key_denoise=DEFAULTS["add_key_denoised"],
    cmap="tab20",
    axs=None,
):
    """Visualise HTO distribution across different stages: raw, normalised, and denoised."""
    assert not (layer_raw == use_key_normalise), (
        "Raw and normalised layers must be different"
    )
    assert not (layer_raw == use_key_denoise), (
        "Raw and denoised layers must be different"
    )

    if axs is None:
        fig, axs = plt.subplots(3, 1, figsize=figsize)
    plt.tight_layout()

    def _plot_temp(
        ax,
        layer,
        title,
        use_log,
    ):
        ax = distribution(
            adata,
            ax=ax,
            layer=layer,
            cmap=cmap,
            title=title,
            fill=True,
            remove_legend=False,
            use_log=use_log,
        )
        return ax

    ax = axs[0]
    ax = _plot_temp(
        ax=ax,
        layer=layer_raw,
        title="Logged Raw Data",
        use_log=True,
    )
    ax.set_xlabel("")

    ax = axs[1]
    ax = _plot_temp(
        ax=ax,
        layer=use_key_normalise,
        title="Normalised Data",
        use_log=False,
    )
    ax.set_xlabel("")

    ax = axs[2]
    ax = _plot_temp(
        ax=ax,
        layer=use_key_denoise,
        title="Denoised Data",
        use_log=False,
    )

    # add lines
    if highlight is not None:
        # get cmap from ax
        cmap = dict(zip(adata.var_names, sns.color_palette(cmap)))
        var = adata.var_names

        ax = axs[0]
        for v in var:
            value = adata[highlight, v].X.data
            value = np.log1p(value)
            ax.axvline(value, c=cmap[v])

        ax = axs[1]
        for v in var:
            value = adata[highlight, v].layers[use_key_normalise]
            ax.axvline(value, c=cmap[v])

        ax = axs[2]
        for v in var:
            value = adata[highlight, v].layers[use_key_denoise]
            ax.axvline(value, c=cmap[v])

    return axs



SUPPORTED_STEPS = ["normalise", "denoise", "demux", "paths"]


def init_meta(adata):
    """Initialize the meta data for the AnnData object.

    DNDs metadata is structured as follows:
    adata.uns["dnd"] = {
        "normalise": {
            "method": "method_name",
            "params": {"param1": value1, "param2": value2, ...},
        },
    """
    initial_meta = {step: {} for step in SUPPORTED_STEPS}
    existing_meta = adata.uns.get("dnd", {})
    adata.uns["dnd"] = {
        **initial_meta,
        **existing_meta,
    }
    return adata


def add_meta(adata, step, params={}, **kwargs):
    """Add metadata to the AnnData object.

    Args:
        adata: AnnData object.
        step: The step of the DND pipeline (should be in "normalise", "denoise", "demux", "paths")
        params: Parameters used in the step.
        kwargs: Additional metadata to add.

    Example:
    ```
    adata.uns["dnd"] = {
        "normalise": {},
        "denoise": {},
        "demux": {},
    }
    adata = add_meta(adata, step="normalise", params={"method": "method_name"}, key1="value1", key2="value2")
    # >> adata.uns["dnd"]["normalise"]["key1"] == "value1"
    # >> adata.uns["dnd"]["normalise"]["params"]["method"] == "method_name"
    ```

    """
    assert step in ["normalise", "denoise", "demux", "paths"], f"Invalid step: '{step}'"
    meta_new = {
        step: {
            "params": params,
            **kwargs,
        },
    }
    adata.uns["dnd"] = {
        **adata.uns["dnd"],
        **meta_new,
    }
    return adata



"""Create a horizontal stacked barplot from a DataFrame."""

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def barplot(
    df: pd.DataFrame,
    by: list,
    labels: list = None,
    order_rows: list = None,
    order_cols: list = None,
    cmap: dict = None,
    title: str = "",
    ax: plt.Axes = None,
    **kwargs,
):
    """Create a horizontal stacked barplot from a DataFrame.

    Args:
        df (pd.DataFrame): DataFrame containing the data to plot.
        by (list): Columns to group by. The first column is used for the x-axis, the second for the y-axis.
        labels (list, optional): Labels for the 'by' columns. Defaults to the column names.
        order_rows (list, optional): Order of rows. Defaults to implicit order.
        order_cols (list, optional): Order of columns. Defaults to implicit order.
        cmap (dict, optional): Color map for the bars. Defaults to None.
        title (str, optional): Title of the plot. Defaults to "".
        ax (plt.Axes, optional): Axes to plot on. If None, a new figure is created. Defaults to None.
        **kwargs: Additional keyword arguments for the seaborn barplot.

    """
    # set kwargs
    kwargs_defaults = {
        "width": 0.8,
        "stacked": True,
    }
    kwargs = {**kwargs_defaults, **kwargs}

    # init
    if not isinstance(by, list):
        by = [by]
    if cmap is None:
        cmap = dict(zip(df[by[0]].unique(), sns.color_palette("tab20")))

    # if x is a list, then make them to long format
    df_long = df[by].melt()
    x = "variable"
    y = "value"

    # rename
    labels = labels or by
    assert len(labels) == len(by), "Labels must match the number of 'by' columns."
    map_labels = dict(zip(by, labels))
    df_long[x] = df_long[x].map(map_labels)

    # aggregate
    df_agg = df_long.groupby([x, y], observed=True).size().unstack()
    df_agg = df_agg / df_agg.sum(axis=1).values[:, None]

    # order
    if order_rows is None:
        order_rows = df_agg.index
    order_rows = [r for r in order_rows if r in df_agg.index]
    order_cols = [c for c in cmap.keys() if c in df_agg.columns]
    df_agg = df_agg.loc[order_rows[::-1], order_cols]

    # colors
    if order_cols is None:
        order_cols = df_agg.columns

    # plot
    if ax is None:
        fig, ax = plt.subplots(figsize=(6, 2))
    ax = df_agg.loc[labels[::-1]].plot.barh(color=cmap, ax=ax, **kwargs)
    ax.set_title(title)
    ax.legend(bbox_to_anchor=(1.4, 1), loc="upper right")

    return ax



"""End-to-end demultiplexing using DND.

1. Step: Preprocessing
2. Step: Normalization
3. Step: Denoising
4. Step: Demultiplexing
5. Step: Postprocessing
"""

import warnings

import anndata as ad

from ._classify import assert_demux
from ._cluster_background import assert_background
from ._defaults import DEFAULTS
from ._logging import get_logger
from ._utils import (
    add_docstring,
    get_arg,
    read_adata,
    read_whitelist,
    subset_whitelist,
    test_write,
    user_input_error_decorator,
    write_csv_safe,
    write_h5ad_safe,
)
from .demux import demux
from .denoise import denoise
from .normalise import normalise
from .report import report_safe

warnings.filterwarnings("ignore", module="anndata")


@user_input_error_decorator
@add_docstring()
def demultiplex(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData = None,
    adata_gex: ad.AnnData = None,
    adata_background: ad.AnnData = None,
    adata_out: str = None,
    csv_out: str = None,
    path_report: str = None,
    show_report: bool = False,
    _as_cli: bool = False,  # required when run as cli
    **kwargs,
):
    """Perform normalization and demultiplexing on the provided filtered and raw AnnData objects.

    Args:
        adata_hto (anndata.AnnData): {adata_hto}
        adata_hto_raw (anndata.AnnData, optional): {adata_hto_raw}
        adata_gex (anndata.AnnData, optional): {adata_gex}
        adata_background (anndata.AnnData, optional): {adata_background}
        adata_out (str, optional): {adata_out}
        csv_out (str, optional): {csv_out}
        path_report (str, optional): {path_report}
        show_report (bool, optional): Show report in current session after processing.
        _as_cli (bool, optional): Run code in CLI mode. This triggers additional checks and logging.
        **kwargs: Allows for keywords such as: `inplace`, `verbose`, `denoise_version`, `add_key_hashid`, `add_key_doublet`, `add_key_normalise`, `add_key_denoised`, `demux_method`, `background_method`, `background_version`

    Returns:
        anndata.AnnData: Anndata object with normalised and denoised expression levels, and sample assignments stored in .obs

    """
    # SET PARAMS
    inplace = get_arg("inplace", kwargs, DEFAULTS)
    verbose = get_arg("verbose", kwargs, DEFAULTS)
    denoise_version = get_arg("denoise_version", kwargs, DEFAULTS)
    add_key_hashid = get_arg("add_key_hashid", kwargs, DEFAULTS)
    add_key_doublet = get_arg("add_key_doublet", kwargs, DEFAULTS)
    add_key_normalise = get_arg("add_key_normalise", kwargs, DEFAULTS)
    add_key_denoised = get_arg("add_key_denoised", kwargs, DEFAULTS)
    demux_method = get_arg("demux_method", kwargs, DEFAULTS)
    background_method = get_arg("background_method", kwargs, DEFAULTS)
    background_version = get_arg("background_version", kwargs, DEFAULTS)
    force = get_arg("force", kwargs, DEFAULTS)

    # LOGGER
    logger = get_logger("dnd", level=verbose)

    # READ
    # Note - adata_hto can also be a whitelist of barcodes
    if isinstance(adata_hto_raw, str):
        logger.debug(f"Reading raw hto adata from {adata_hto_raw}")
        adata_hto_raw = read_adata(adata_hto_raw)

    if isinstance(adata_hto, str):
        if adata_hto.endswith(".h5ad"):
            logger.debug(f"Reading filtered adata from {adata_hto}")
            adata_hto = read_adata(adata_hto)
        elif adata_hto.endswith(".csv") or adata_hto.endswith(".csv.gz"):
            logger.debug(f"Reading whitelist from {adata_hto}")
            whitelist = read_whitelist(adata_hto)
            adata_hto = subset_whitelist(adata_hto_raw, whitelist)
        else:
            msg = f"Unknown file format for adata_hto: {adata_hto}. Must be anndata (.h5ad) or whitelist (.csv|.csv.gz)"
            raise ValueError(msg)

    if isinstance(adata_background, str):
        logger.debug(f"Reading background adata from {adata_background}")
        adata_background = read_adata(adata_background)

    # ASSERTIONS
    # - check that output path is writeable (and .h5ad)
    # - check that parameters are valid
    # - check that keys do not exist in adata if run as cli
    test_write(adata_out, filetype="h5ad", create_folder=True, _require_write=_as_cli)
    test_write(csv_out, filetype="csv", create_folder=True, _require_write=_as_cli)
    test_write(path_report, filetype="pdf", create_folder=True, _require_write=_as_cli)
    assert_background(background_method)
    assert_demux(demux_method)
    if _as_cli:
        assert adata_out is not None, (
            "Output path must be provided using parameter --output-path"
        )
        assert (force) or (add_key_normalise not in adata_hto.layers), (
            f"Key {add_key_normalise} already exists in adata. Add option --add-key-normalise to change the key."
        )
        assert (force) or (add_key_denoised not in adata_hto.layers), (
            f"Key {add_key_denoised} already exists in adata. Add option --add-key-denoise to change the key."
        )

    # WARNINGS
    if inplace:
        logger.warning(
            "Behaviour with `inplace=True` works, but is not recommended. Some behaviours may change in future versions."
        )

    # GET DATA
    if (
        adata_gex is not None
        and isinstance(adata_gex, str)
        and background_version in ["v1", "v3"]
    ):
        logger.debug(f"Reading gex adata from {adata_gex}")
        adata_gex = read_adata(adata_gex)

    # LOG
    logger.info(
        f"Starting DND: Normalise (build-background: {background_version}) -> Denoise (background-dection: {background_method} | version: {denoise_version}) -> Demux (method: {demux_method})"
    )

    # RUN
    adata_hto = normalise(
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        adata_background=adata_background,
        background_version=background_version,
        inplace=inplace,
        verbose=verbose,
        add_key_normalise=add_key_normalise,
        use_layer=get_arg("use_layer", kwargs, DEFAULTS),
        pseudocount=get_arg("pseudocount", kwargs, DEFAULTS),
        k_gex_cells=get_arg("k_gex_cells", kwargs, DEFAULTS),
        next_k_cells=get_arg("next_k_cells", kwargs, DEFAULTS),
        min_umi=get_arg("min_umi", kwargs, DEFAULTS),
    )

    adata_hto = denoise(
        adata_hto=adata_hto,
        inplace=inplace,
        verbose=verbose,
        use_layer=add_key_normalise,
        add_key_denoised=add_key_denoised,
        background_method=background_method,
        denoise_version=denoise_version,
        covariates=get_arg("covariates", kwargs, DEFAULTS),
        design=get_arg("design", kwargs, DEFAULTS),
        kwargs_denoise=get_arg("kwargs_denoise", kwargs, DEFAULTS),
    )

    adata_hto = demux(
        adata_hto=adata_hto,
        inplace=inplace,
        verbose=verbose,
        use_layer=add_key_denoised,
        demux_method=demux_method,
        key_normalise=add_key_normalise,
        enforce_larger_than_background=get_arg(
            "enforce_larger_than_background", kwargs, DEFAULTS
        ),
        add_key_hashid=add_key_hashid,
        add_key_doublet=add_key_doublet,
        add_key_labels=get_arg("add_key_labels", kwargs, DEFAULTS),
        kwargs_classify=get_arg("kwargs_classify", kwargs, DEFAULTS),
    )

    # SAVE
    write_h5ad_safe(adata_hto, adata_out, create_folder=True, _require_write=_as_cli)
    write_csv_safe(
        adata_hto,
        csv_out,
        key_hashid=add_key_hashid,
        key_doublet=add_key_doublet,
        create_folder=True,
    )

    # REPORT (not yet tested well)
    report_safe(
        adata_hto=adata_hto,
        adata_background=adata_background,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        path_report=path_report,
        use_key_normalise=add_key_normalise,
        use_key_denoise=add_key_denoised,
        show=show_report,
        verbose=verbose,
    )

    return adata_hto



import numpy as np
from skimage._shared.utils import warn
from skimage.filters.thresholding import _validate_image_histogram


def threshold_otsu_weighted(
    image=None, nbins=256, p_target=None, lam=1.0, *, hist=None
) -> float:
    """Weighted implementation of sklearn's threshold_otsu function.

    Source code from: https://github.com/scikit-image/scikit-image/blob/v0.25.2/skimage/filters/thresholding.py#L336-L407

    Changes:
    - Added a soft prior (p_target) which penalizes deviations of the positive class and pushes the threshold accordingly.
    - Added lam parameter to control the strength of the penalty.
    """
    if p_target is None:
        raise ValueError("'p_target' must be provided for weighted Otsu thresholding.")

    if image is not None and image.ndim > 2 and image.shape[-1] in (3, 4):
        warn(
            f"threshold_otsu is expected to work correctly only for "
            f"grayscale images; image shape {image.shape} looks like "
            f"that of an RGB image."
        )

    # Check if the image has more than one intensity value; if not, return that
    # value
    if image is not None:
        first_pixel = image.reshape(-1)[0]
        if np.all(image == first_pixel):
            return first_pixel

    counts, bin_centers = _validate_image_histogram(image, hist, nbins)

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]

    # class means for all possible thresholds
    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of ``weight1``/``mean1`` should pair with zero values in
    # ``weight2``/``mean2``, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2
    variance_norm = variance12 / (variance12.max() + 1e-5)

    # normalize weight2 to get actual proportions (0-1 scale)
    total_counts = weight1[-1]  # total number of pixels/points
    pos_frac = weight2[1:] / total_counts

    # penalise deviations from target
    prior = np.exp(-lam * (pos_frac - p_target) ** 2)
    score = variance_norm * prior
    idx = np.argmax(score)

    threshold = bin_centers[idx]

    return threshold
